/*
 *    LocBoost2.java
 *    Copyright (C) 2000 Gilad Mishne
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package weka.classifiers;

import java.io.*;
import java.util.*;
import weka.core.*;
import weka.optimization.*;

/**
 * Class for boosting a classifier using LocBoost2.
 * For more information, see<p>
 *
 * Ron Meir, Ran El-Yaniv, Shai Ben-David
 * (2000). <i>Localized Boosting</i>.
 * insert-here: where this paper will appear<p>
 *
 * Valid options are:<p>
 *
 * -D <br>
 * Turn on debugging output.<p>
 *
 * -W classname <br>
 * Specify the full class name of a classifier as the basis for 
 * boosting (required).<p>
 *
 * -I num <br>
 * Set the number of boost iterations (default 10). <p>
 *
 * Options after -- are passed to the designated classifier.<p>
 *
 * @author Gilad Mishne (gilad.mishne@intel.com)
 * @version $Revision: 1.0 $ 
 */
public class LocBoost2 extends DistributionClassifier
  implements OptionHandler, WeightedInstancesHandler, Sourcable {

  public class Phi {
    public double meanX;
    public double meanY;
    public double sigma;
  }
  
  /**
   * Main Q Function and its gradients
   */
  public class Q_phi implements MinFuncInterface {
    
    private Phi phic;
    private double[] thetac;

    public Q_phi(Phi phic_, double[] thetac_) {
        phic = phic_;
        thetac = thetac_;
    }    
    
    public double f(double x[]) {
        Phi phi = new Phi();
        phi.meanX = x[0];
        phi.meanY = x[1];
        phi.sigma = x[2];

        double res = 0;        
        double gam,ht;
        
        Enumeration enumInsts = m_Instances.enumerateInstances();
        while (enumInsts.hasMoreElements()) {
            Instance inst = (Instance)enumInsts.nextElement();
            gam = gamma(inst, phi);
            if (gam<=0)
                gam=eps;
            if (gam>=1)
                gam=1-eps;
            ht = ht_1(phic,thetac,inst);
            res += ht*Math.log(gam) + (1-ht)*Math.log(1-gam);
        }
        return (-res);  // return (-res) instead of res because optimization minimizes        
    }
    
    public double[] grad(double x[]) {
        double[] res = new double[x.length];
        
        // create gradient numerically - for debug
        for (int i=0; i<x.length; i++)
            res[i] = grad_numeric(x,i);          
        
        return ArrUtils.arr_mul(res, -1);   // return -res because f is reversed
    }
    
    protected double grad_numeric(double[] x, int var) {
        double epsilon = x[var]==0?1e-3:x[var]/1e3;
        double[] newvalues = (double[])x.clone();
        newvalues[var] += epsilon;
        return (f(x)-f(newvalues))/epsilon;        
    }
  }
  
  public class Q_theta implements MinFuncInterface {
    
    private Phi phic;
    private double[] thetac;

    public Q_theta(Phi phic_, double[] thetac_) {
        phic = phic_;
        thetac = thetac_;
    }    
    
    public double f(double x[]) {
        double[] theta = (double[])x.clone();
        double res = 0;        
        double p,ht;
        
        Enumeration enumInsts = m_Instances.enumerateInstances();
        while (enumInsts.hasMoreElements()) {
            Instance inst = (Instance)enumInsts.nextElement();
            int class_ = inst.value(2)>0?1:-1;
            p = p(inst, theta, class_);
            if (0>=p)
                p=eps;
            ht = ht_1(phic,thetac,inst);
            res += ht*Math.log(p);
        }
        return (-res);  //    return (-res) instead of res because optimization minimizes        
    }
    
    public double[] grad(double x[]) {
        double[] res = new double[x.length];
        
        // create gradient numerically - for debug
        for (int i=0; i<x.length; i++)
            res[i] = grad_numeric(x,i);          
        
        return ArrUtils.arr_mul(res, -1);   // return -res because f is reversed
    }
    
    protected double grad_numeric(double[] x, int var) {
        double epsilon = x[var]==0?1e-3:x[var]/1e3;
        double[] newvalues = (double[])x.clone();
        newvalues[var] = x[var] + epsilon;
        return (f(x)-f(newvalues))/epsilon;        
    }        
  }
  
  /** Enum for initializing parameters method */ 
  private final static double eps = 1e-6;
  
  /** Arrays for storing the weights for the votes. */
  protected Phi [] m_phis;
  protected double [][] m_thetas;

  /** The maximum number of boost iterations */
  protected int m_MaxIterations;

  /** The number of successfully generated base classifiers. */
  public int m_NumIterations;

  /** Debugging mode, gives extra output if true */
  protected boolean m_Debug = true; 
  
  /** Use optimization for theta, phi */
  protected boolean m_Optimize = true; 

  /** Overfit - try to classify incorrect groups of size 1 */
  protected boolean m_Overfit = false; 

  /** Number of iteration in the Quasi-Newton optimization 1 */
  protected int m_OptimSteps = 2; 

  /** The number of classes */
  protected int m_NumClasses;
  
/*******************************/

  /** instances for building classifier */
  protected Instances m_Instances;
  
  /** For the LocBoost2 applet */
  public double[] m_CenterOfUnknownGroup = new double[2];
  /** For the LocBoost2 applet */
  public double m_RadiusOfUnknownGroup;

/*******************************/


  /**
   * Returns an enumeration describing the available options
   *
   * @return an enumeration of all the available options
   */
  public Enumeration listOptions() {
    return null;
  }


  /**
   * Parses a given list of options. Valid options are:<p>
   *
   * -D <br>
   * Turn on debugging output.<p>
   *
   * -W classname <br>
   * Specify the full class name of a classifier as the basis for 
   * boosting (required).<p>
   *
   * -I num <br>
   * Set the number of boost iterations (default 10). <p>
   *
   * -P num <br>
   * Set the percentage of weight mass used to build classifiers
   * (default 100). <p>
   *
   * -Q <br>
   * Use resampling instead of reweighting.<p>
   *
   * -S seed <br>
   * Random number seed for resampling (default 1).<p>
   *
   * Options after -- are passed to the designated classifier.<p>
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {
  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] options = new String [0];
    return options;
  }

  /**
   * Set the maximum number of boost iterations
   */
  public void setMaxIterations(int maxIterations) {

    m_MaxIterations = maxIterations;
  }

  /**
   * Get the maximum number of boost iterations
   *
   * @return the maximum number of boost iterations
   */
  public int getMaxIterations() {
    return m_MaxIterations;
  }

  /**
   * Set debugging mode
   *
   * @param debug true if debug output should be printed
   */
  public void setDebug(boolean debug) {
    m_Debug = debug;
  }

  /**
   * Get whether debugging is turned on
   *
   * @return true if debugging output is on
   */
  public boolean getDebug() {
    return m_Debug;
  }

  /**
  * Simple Logistic function.
  * @param Instance, thetas, value of y
  * @return 1/1+exp(-y*(thetas DOT_PRODUCT instance))
  */
  public double p(Instance instance, double[] theta, int class_) { 
      double v = theta[2];
      for (int i=0; i<2; i++) {    // dot product
          v += instance.value(i)*theta[i];
      }
      double tmp = 1+Math.exp(-v*class_);
      if (0==tmp)
        tmp=eps;
      return 1/tmp;  
  }

  /**
   * Calculate the weight of a classifier regarding an instance.
   *
   * @param Instance, localized weight parameter (phi)
   * @return weight of the classifier when predicting the instance (beta)
   */
  protected double beta(Instance instance, Phi phi) { 
    // row is 0 because x.y are mutually independent (??)
    double x = instance.value(0);
    double y = instance.value(1);
    double relativeDistance = 0;
    relativeDistance += Math.pow((x-phi.meanX),2);
    relativeDistance += Math.pow((y-phi.meanY),2);
    relativeDistance = Math.sqrt(relativeDistance);
    double tmp = (phi.sigma!=0)?phi.sigma:eps;
    relativeDistance /= tmp;
    return Math.exp(-0.5*relativeDistance);
  }

  /**
   * multivariate normal distribution.
   *
   * @param Instance, localized weight parameter (phi)
   * @return weight of this instance regarding phi
   */
  protected double gamma(Instance instance, Phi phi) { 
    // row is 0 because x.y are mutually independent (??)
    double x = instance.value(0);
    double y = instance.value(1);
    double sumOfSquares = Math.pow((x-phi.meanX),2) + Math.pow((y-phi.meanY),2);
    double tmp = (phi.sigma!=0)?phi.sigma:eps;
    return Math.exp(-0.5*(sumOfSquares/tmp));
  }

  protected double ht_1(Phi phic, double[] thetac, Instance inst) {
      double gam = gamma(inst, phic);
      int class_ = inst.value(2)>0?1:-1;
      double p = p(inst, thetac, class_);
      double tmp = (1-gam)*P(m_NumIterations-1,inst, class_) + gam*p;
      if (0==tmp)
        tmp = eps; 
      return  (gam*p) / tmp;        
  }
  
  public void setOptimization(boolean state) {
    m_Optimize = state;
  }
  public void setOverfitting(boolean state) {
    m_Overfit = state;
  }
  public void setOptimSteps(int steps) {
    m_OptimSteps = steps;
  }
  public boolean getOptimization() {
    return m_Optimize;
  }
  public boolean getOverfitting() {
    return m_Overfit;
  }
  public int getOptimSteps() {
    return m_OptimSteps;
  }
  
  /**
   * Perform EM-algorithm on Q, on phi and theta seperately
   */
  protected void maximizeQFunction() { 
    //if (m_NumIterations==0) return;
    int t = m_NumIterations;
    Phi phic = m_phis[t];
    double[] thetac = m_thetas[t];
    Q_phi qp = new Q_phi(phic, thetac );
    Q_theta qt = new Q_theta(phic, thetac);
    double[] initValuesP = new double[3];
    double[] initValuesT = new double[3];
    double[] optimisedValuesP = new double[3];
    double[] optimisedValuesT = new double[3];
    initValuesP[0] = phic.meanX;
    initValuesP[1] = phic.meanY;
    initValuesP[2] = phic.sigma;
    initValuesT[0] = thetac[0];
    initValuesT[1] = thetac[1];
    initValuesT[2] = thetac[2];

    if (m_Optimize) {
        optimisedValuesT = QuasiNewton.optimize(3, initValuesT, qt, 3);
        optimisedValuesP = QuasiNewton.optimize(3, initValuesP, qp, 3);
    } else {
        for (int p=0; p<3;p++) {
            optimisedValuesP[p] = initValuesP[p];
            optimisedValuesT[p] = initValuesT[p];
        }
    }

    // update values
    m_phis[t].meanX = optimisedValuesP[0];
    m_phis[t].meanY = optimisedValuesP[1];
    m_phis[t].sigma = optimisedValuesP[2];
    m_thetas[t][0] = optimisedValuesT[0];
    m_thetas[t][1] = optimisedValuesT[1];
    m_thetas[t][2] = optimisedValuesT[2];
    
  }

  /**
   * Get current model.
   */
  protected double P(int t, Instance x, int class_) {
    if (0>=t) return 0.5;       // recursion end
    double gam = gamma(x, m_phis[t]);
    return (1-gam)*P(t-1, x, class_) + gam*p(x, m_thetas[t], class_);
  }


  /**
   * Mark set of instances for which the classifier is wrong.
   */
  protected int markPoorlyIdentifiedInstances() { 
    int numOfPoors = 0;
    Enumeration enum = m_Instances.enumerateInstances();
    while (enum.hasMoreElements()) {
	    Instance instance = (Instance)enum.nextElement();
	    instance.unmarkInstance();
	    double p = P(m_NumIterations-1, instance, -1);
	    int class_ = p<0.5?1:-1;
	    int realClass = instance.value(2)==0?-1:1;
        if (class_ != realClass ) {
	        instance.markInstance();
	        numOfPoors++;
	    }
    }
    return numOfPoors;
  }

  /**
   * Build graph from instances.
   */
  protected Graph createGraphFromData(Instances V) {
        
    if (V.numInstances()<1) return null;
    
    Graph g = new Graph();
    Enumeration Venum = V.enumerateInstances();
    while (Venum.hasMoreElements()) {
        g.newVertex((Instance)Venum.nextElement());
    }
    Instance tmpInstTo, tmpInstFrom;
    double minDistance, tmp;
    Enumeration Senum = V.enumerateInstances();
    int i=0;
    int modFactor = 1+(V.numInstances()/100);// for spacing of iterations        
    while (Senum.hasMoreElements()) {
        tmpInstFrom = (Instance)Senum.nextElement();
        if (!tmpInstFrom.isMarked()) continue;  // only enum over elements in "S"
        if ((i++)%modFactor!=0) continue;  
        
        // find minimum distance such that an instance is in that distance 
        // from a vertex, but the instance is not in S
        minDistance = Double.MAX_VALUE;
        Venum = V.enumerateInstances();
        while (Venum.hasMoreElements()) {
            tmpInstTo = (Instance)Venum.nextElement();
            if ( (!tmpInstTo.isMarked()) &&
                 ((tmp = tmpInstFrom.distanceTo(tmpInstTo)) < minDistance) ) {
                    minDistance = tmp;
                 }
            if (minDistance<0.05) break;       // performance ...
        }
        // now add an edge between the vertices to the graph
        Venum = V.enumerateInstances();
        Graph.Vertex from = g.getVertex(tmpInstFrom);                
        while (Venum.hasMoreElements()) {
            tmpInstTo = (Instance)Venum.nextElement();
            if (tmpInstFrom.distanceTo(tmpInstTo) < minDistance) {
                Graph.Vertex to = g.getVertex(tmpInstTo);         // save on hash access
                g.newEdge(to, from);
                g.newEdge(from, to);
            }
        }
    }
    return g;
  }
  
  /**
   * Perform an iteration of the LocBoost2 algorithm.
   */
  public boolean iterate() { 
    if (initParams()==0) {       // "(a)"
        this.m_NumIterations--;
        return false;
    } else
        maximizeQFunction();       // "(b)", "(c)"
    return true;
  }

  /**
   * Find initial values for theta and phi.
   *
   * @param method the method to use for the initial value guess.
   */
  protected int initParams() {
        
    int t = 0;
    if (markPoorlyIdentifiedInstances()==0) {
        return 0;
    }            
    Graph g  = createGraphFromData(m_Instances);
    g.findSCCs();
    Vector sccVec = g.getLargestSCC();
    int sccSize = sccVec.size();
    if (sccSize==0)
        return 0;
    if (sccSize==1) {
        if (m_Overfit) {
            m_CenterOfUnknownGroup[0]= ((Instance)sccVec.elementAt(0)).value(0);
            m_CenterOfUnknownGroup[1]= ((Instance)sccVec.elementAt(0)).value(1);
            m_RadiusOfUnknownGroup = 0.0005;
        } else
            return 0;
    } else {
        double sumX=0;
        double sumY=0;
        for (int i=0 ; i< sccSize ; i++) {
            sumX += ((Instance)sccVec.elementAt(i)).value(0);
            sumY += ((Instance)sccVec.elementAt(i)).value(1);
        }
        m_CenterOfUnknownGroup[0]= (sccSize>0)?sumX/sccSize:0;
        m_CenterOfUnknownGroup[1]= (sccSize>0)?sumY/sccSize:0;
                
        double e_x2=0;
        double ex_2=0;
        double disX, disY;
        for (int i=0 ; i< sccSize ; i++) {
            disX = ((Instance)sccVec.elementAt(i)).value(0)-m_CenterOfUnknownGroup[0];
            disY = ((Instance)sccVec.elementAt(i)).value(1)-m_CenterOfUnknownGroup[1];
            e_x2 += disX*disX + disY*disY;
            ex_2 += disX + disY;
        }
        if (sccSize>0) {
            e_x2 /= sccSize;
            ex_2 /= sccSize;
        }
        ex_2 *= ex_2;
        double sigma = e_x2-ex_2;
        m_RadiusOfUnknownGroup = sigma;
    }
            
    // update values
    m_phis[m_NumIterations] = new Phi();
    m_phis[m_NumIterations].meanX = m_CenterOfUnknownGroup[0];
    m_phis[m_NumIterations].meanY = m_CenterOfUnknownGroup[1];
    m_phis[m_NumIterations].sigma = m_RadiusOfUnknownGroup;

    Instances tmpInsts = new Instances(m_Instances, sccSize);
    for (int i=0 ; i< sccSize ; i++) {
        tmpInsts.add((Instance)sccVec.elementAt(i));
    }            
    Logistic tmpL = new Logistic();
    try {
        tmpL.buildClassifier(tmpInsts);
    } catch (Exception e) { e.printStackTrace(); }                
    m_thetas[m_NumIterations] = new double[3];
    m_thetas[m_NumIterations][0] = tmpL.m_Par[1]; 
    m_thetas[m_NumIterations][1] = tmpL.m_Par[2];            
    m_thetas[m_NumIterations][2] = tmpL.m_Par[0];                              
    return 1;
  }    
            
  /**
   * Boosting method.
   *
   * @param data the training data to be used for generating the
   * boosted classifier.
   * @exception Exception if the classifier could not be built successfully
   */

  public void buildClassifier(Instances data) throws Exception {
    
    data.deleteWithMissingClass();
    if (data.numInstances() == 0) {
      throw new Exception("No good train instances !");
    }
    
    if (data.classAttribute().isNumeric() || data.numClasses() !=2) {
      throw new Exception("Sorry, LocBoost2 supports only 2-class nominal datasets currently.");
    }
    
    m_Instances = data;
    m_phis = new Phi[500];   //bug bug bug
    m_thetas = new double[500][3];  //bug bug bug
    m_NumIterations=0;
    
    // later, enable boosting up to a point where some accuracy is reached
    // rather than a fixed number of times T
    
    while (m_NumIterations<m_MaxIterations) {
        m_NumIterations++;
        iterate();
    }
  }

  /**
   * Calculates the class membership probabilities for the given test instance.
   *
   * @param instance the instance to be classified
   * @return predicted class probability distribution
   * @exception Exception if instance could not be classified
   * successfully
   */
  public double [] distributionForInstance(Instance instance) 
    throws Exception {
        
    double [] sums = new double [2]; 
      
    sums[0] = P(m_NumIterations, instance, -1);
    sums[1] = 1- sums[0];
       
    return sums;
  }

  /**
   * Returns the boosted model as Java source code.
   *
   * @return the tree as Java source code
   * @exception Exception if something goes wrong
   */
  public String toSource(String className) throws Exception {
    return new String("");
  }

  /**
   * Returns description of the boosted classifier.
   *
   * @return description of the boosted classifier as a string
   */
  public String toString() {
    StringBuffer res = new StringBuffer();
    res.append("LocBoost - Localized Booster ( Logistic-optimized version),\nrunning for " + m_NumIterations + " iterations.\n");
    res.append("=========================================================\n\n");
    for (int i=0; i<m_NumIterations; i++) {
        int realI = i+1;
        res.append("Iteration number " + realI + ":\n");
        res.append("---------------------------------------------------------\n");
        res.append("Weak classifier is centered at  ( " + m_phis[realI].meanX + " , " + m_phis[realI].meanY + " ),\n");
        res.append("with a radius of " + m_phis[realI].sigma + ":\n\n");
        res.append("Logistic coefficients:\n");
        res.append(m_thetas[realI][0] + ",  "+ m_thetas[realI][1] + ",  " + m_thetas[realI][2]);
        res.append("\n\n");
    }        
    return res.toString();
  }

  /**
   * Main method for testing this class.
   *
   * @param argv the options
   */
  public static void main(String [] argv) {
    try {
      System.out.println(Evaluation.evaluateModel(new LocBoost2(), argv));
    } catch (Exception e) {
      System.err.println(e.getMessage());
    }
  }
}

