/*
 *    QuasiNewton.java
 *    Copyright (C) 2000 Gilad Mishne
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  
 *    This file is based on the Matlab Quasi-Newton code distributed with Netlab.
 *    It was translated to Java by the author listed below.
 *
 * @author Gilad Mishne (gilad.mishne@intel.com)
 * @version $Revision: 1.0 $ 
 */

package weka.optimization;

import weka.optimization.*;

/**
*
*<p>
*This class contains Java translations of the UNCMIN unconstrained 
*optimization routines.  See R.B. Schnabel, J.E. Koontz, and B.E.
*Weiss, <i>A Modular System
*of Algorithms for Unconstrained Minimization</i>, Report CU-CS-240-82,
*Comp. Sci. Dept., University of Colorado at Boulder, 1982.
*
*/

public class QuasiNewton extends Object {

/**
*
*
*@param  n         The number of arguments of the function to minimize
*@param  x         The initial estimate of the minimum point
*@param minClass   A class that implements the Uncmin_methods
*@param maxIter    Maximum number of iterations. If maxIter < 1, 100 is taken.
*
*/

   public static double[] optimize(int n, double x[], MinFuncInterface minClass,
                                 int maxIter) {

        double fnew, fold, lmin, vGv, vdotp;
        double[] gradnew, gradold, sd;
        double[][] hessinv;
        double[]  xold, v, Gv;

        if(maxIter<1)
            maxIter=100;

        final double eps = 1.0e-8;
        final double precision2 =  1e-3;  // Default
        final double precision3 =  1e-4;  // Default

        // Minimal fractional change in f from Newton step: otherwise do a line search
        final double min_frac_change = 1e-3;

        fnew = minClass.f(x);
        if (Double.isNaN(fnew)) {
            System.out.println("QuasiNewton: The initial values supplied yield a NaN !");
            return x;
        }
        gradnew = minClass.grad(x);
        sd = ArrUtils.arr_mul(gradnew,-1);		// Search direction
        hessinv = ArrUtils.eye(n);               // Initialize inverse Hessian to be identity matrix
        int j = 0;
        while (j < maxIter) {
            xold = ArrUtils.arr(x);
            fold = fnew;
            gradold = ArrUtils.arr(gradnew);

            x = ArrUtils.arr_plus(xold,sd);
            fnew = minClass.f(x);
            if (Double.isNaN(fnew)) {
                System.out.println("QuasiNewton: The values reached yield a NaN !");
                return x;
            }

            // This shouldn't occur, but rest of code depends on sd being downhill
            if (ArrUtils.arr_mul(gradnew,sd) >= 0) {
                sd = ArrUtils.arr_mul(sd,-1);
            }


            // Does the Newton step reduce the function value sufficiently?
            if ((fnew >= fold + (min_frac_change*(ArrUtils.arr_mul(gradnew,sd))))) {
                // No it doesn't
                // Minimize along current search direction: must be less than Newton step
                lmin = linemin(minClass, xold, sd, fold, -1, n, precision2);
                // Correct x and fnew to be the actual search point we have found
                x = ArrUtils.arr_plus(xold,ArrUtils.arr_mul(sd,lmin));
                sd = ArrUtils.arr_minus(x,xold);
                fnew = minClass.f(x);
            }

            // Check for termination
            if ( ((ArrUtils.arr_max(ArrUtils.arr_abs(ArrUtils.arr_minus(x,xold)))) < precision2 )&& 
                   (Math.abs(fnew-fold) < precision3 ) ) {
                    return x;
                   }

            gradnew = minClass.grad(x);
            v = ArrUtils.arr_minus(gradnew,gradold);
            vdotp = ArrUtils.arr_mul(v,sd);
            if (0==vdotp)
                vdotp = eps;

            // Skip update to inverse Hessian if fac not sufficiently positive
            if ( vdotp*vdotp > eps*ArrUtils.arr_sum(ArrUtils.arr_sqr(v))*ArrUtils.arr_sum(ArrUtils.arr_sqr(sd)) ) {
                Gv = ArrUtils.arr_mul(hessinv,v);
                vGv = ArrUtils.arr_sum(ArrUtils.arr_elem_mul(v,Gv));
                if (0==vGv)
                    vGv = eps;
                v = ArrUtils.arr_minus(ArrUtils.arr_mul(sd,1/vdotp),ArrUtils.arr_mul(Gv,1/vGv));
                // Use BFGS update rule
                hessinv = ArrUtils.mat_plus(hessinv,((ArrUtils.arr_mul(sd,sd))/vdotp)-
                          ((ArrUtils.arr_mul(Gv,Gv))/vGv)+(vGv*(ArrUtils.arr_mul(v,v))));
            }

            sd = ArrUtils.arr_mul(ArrUtils.arr_mul(hessinv,gradnew),-1);

            //System.out.print("Cycle " + j + "  Function: " + fnew + "  Values: ");
            //ArrUtils.arr_print(x);
            j++;

        // If we get here, then we haven't terminated in the given number of iterations.
        //        System.out.println("Warning: Maximum number of iterations has been exceeded in QuasiNewton");
        }
        return x;
    }

    public static double sign(double d) {
        return (d>=0?1.0:-1.0);
    }
    

/**
 *  MINBRACK Bracket a minimum of a function of one variable.
 *  
 *  Description
 *  BRMIN, BRMID, BRMAX, NUMEVALS] = MINBRACK(F, A, B, FA) finds a
 *  bracket of three points around a local minimum of F.  The function F
 *  must have a one dimensional domain. A < B is an initial guess at the
 *  minimum and maximum points of a bracket, but MINBRACK will search
 *  outside this interval if necessary. The bracket consists of three
 *  points (in increasing order) such that F(BRMID) < F(BRMIN) and
 *  F(BRMID) < F(BRMAX). FA is the value of the function at A: it is
 *  included to avoid unnecessary function evaluations in the
 *  optimization routines. The return value NUMEVALS is the number of
 *  function evaluations in MINBRACK.
 *  @param  n   size of arrays
 */
 
    public static double[] minbrack(MinSingleVarFuncInterface minClass, double a, double b) {
     
        double[] res = new double[3];
        boolean bracket_found;
        // Value of golden section (1 + sqrt(5))/2.0
        final double phi = 1.6180339887499;
        // A small non-zero number to avoid dividing by zero in quadratic interpolation
        final double TINY = 1.e-10;
        // Maximal proportional step to take: don't want to make this too big
        // as then spend a lot of time finding the minimum inside the bracket
        double max_step = 10.0;
     
        double fa = minClass.f(a);
        double fb = minClass.f(b);
        double c, r, q, u, ulimit, fc, fu;
        
        // Assume that we know going from a to b is downhill initially 
        // (usually because gradf(a) < 0).
        
        if (fb > fa) {
            // Minimum must lie between a and b: do golden section until we find point 
            // low enough to be middle of bracket
            c = b;
            b = a + (c-a)/phi;
            fb = minClass.f(b);
            while (fb > fa) {
                c = b;
                b = a + (c-a)/phi;
                fb = minClass.f(b);
            }
        } else {
            // There is a valid bracket upper bound greater than b
            c = b + phi*(b-a);
            fc = minClass.f(c);
            bracket_found = false;
            while (fb > fc) {
                // Do a quadratic interpolation (i.e. to minimum of quadratic)
                r = (b-a)*(fb-fc);                
                q = (b-c)*(fb-fa);
                double tmp = 2.0*(sign(q-r)*Math.max(Math.abs(q-r), TINY));
                if (0==tmp)
                    tmp = TINY;
                u = b - ((b-c)*q - (b-a)*r)/TINY;
                fu = minClass.f(u);
                ulimit = b + max_step*(c-b);
                if ((b-u)*(u-c) > 0.0) {
                    // Interpolant lies between b and c
                    fu = minClass.f(u);
                    if (fu < fc) {
	                    // Have a minimum between b and c
	                    res[0] = b;
	                    res[1] = u;
	                    res[2] = c;
	                    return res;
	                } else if (fu > fb) {
	                    // Have a minimum between a and u
	                    res[0] = a;
	                    res[1] = c;
	                    res[2] = u;
	                    return res;
                    }
                    // Quadratic interpolation didn't give a bracket, so take a golden step
                    u = c + phi*(c-b);
                } else if ((c-u)*(u-ulimit) > 0.0) {
                    // Interpolant lies between c and limit
                    fu = minClass.f(u);
                    if (fu < fc) {
	                    // Move bracket along, and then take a golden section step
	                    b = c;
	                    c = u;
	                    u = c + phi*(c-b);
                    } else {
  	                    bracket_found = true;
  	                }
                } else if ((u-ulimit)*(ulimit-c) >= 0.0) {
                    // Limit parabolic u to maximum value
                    u = ulimit;
                    } else {
                        // Reject parabolic u and use golden section step
                        u = c + phi*(c-b);
                    }
                    if (!bracket_found) {
                        fu = minClass.f(u);
                    }
                    a = b;
                    b = c;
                    c = u;
                    fa = fb;
                    fb = fc;
                    fc = fu;
                } // while loop
            }   // bracket found
            res[1] = b;
            if (a < c) {
                res[0] = a;
                res[2] = c;
            } else {
                res[0] = c; 
                res[2] = a;                    
            }
        return res;    
    }

    
    public static double linemin(MinFuncInterface f, double[] pt,
                               double[] dir, double fpt, int maxIter, int n, double precision) {

        if(maxIter<1)
            maxIter=100;
        // Value of golden section (1 + sqrt(5))/2.0
        final double phi = 1.6180339887499;
        final double cphi = 1 - 1/phi;
        final double TINY = 1.0e-10; // Can't use fractional precision when minimum is at 0
        final double TOL = 1.0e-8;	// Maximal fractional precision
        
        double x,v,w,u;
        double d,p,q,tol1,e,xm,r,fx,fv,fu,fw; 
        double br_min, br_mid, br_max;
        xm=0; tol1=0; d=0;   // get rid of annoying java compilation messages
        
        MinSingleVarFuncInterface linef = new Linef();
        linef.setN(n);
        linef.setFunc(f);
        linef.setX(pt);
        linef.setD(dir);
        
        double [] brack = minbrack(linef, 0.0, 1.0);
        br_min = brack[0];
        br_mid = brack[1];
        br_max = brack[2];
        // Use Brent's algorithm to find minimum
        // Initialise the points and function values
        w = br_mid;   	// Where second from minimum is
        v = br_mid;   	// Previous value of w
        x = v;   	        // Where current minimum is
        e = 0.0; 	        // Distance moved on step before last
        
        fx = linef.f(x);
        fv = fx;
        fw = fx;

        for (int n_=0; n_< maxIter; n_++) {
            xm = 0.5*(br_max+br_min);  // Middle of bracket
            // Make sure that tolerance is big enough
            tol1 = TOL * Math.abs(x) + TINY;
            // Decide termination on absolute precision required by options(2)
            if ((Math.abs(x-xm)<=precision) && (br_max-br_min < 4*precision))
                return x;
        }
        // Check if step before last was big enough to try a parabolic step.
        // Note that this will fail on first iteration, which must be a golden section step.
        if (Math.abs(e) > tol1) {
            // Construct a trial parabolic fit through x, v and w
            r = (fx - fv) * (x - w);
            q = (fx - fw) * (x - v);
            p = (x - v)*q - (x - w)*r;
            q = 2.0*(q - r);
            if (q > 0.0) 
                p = -p; 
            q = Math.abs(q);
            // Test if the parabolic fit is OK
            if (Math.abs(p) >= Math.abs(0.5*q*e) || p <= q*(br_min-x) || p >= q*(br_max-x)) {
                // No it isn't, so take a golden section step
                if (x >= xm) {
                    e = br_min-x;
                } else {
                    e = br_max-x;
                }
                d = cphi*e;
            } else {
                // Yes it is, so take the parabolic step
                e = d;
                if (0==q)
                    q=0.001;
                d = p/q;
                u = x+d;
                if ((u-br_min < 2*tol1) || (br_max-u < 2*tol1))
                    d = sign(xm-x)*tol1;
            }
        } else {
            // Step before last not big enough, so take a golden section step
            if (x >= xm)
                e = br_min - x;
            else
                e = br_max - x;
            d = cphi*e;
        }
        // Make sure that step is big enough
        if (Math.abs(d) >= tol1)
            u = x+d;
        else
            u = x + sign(d)*tol1;
        // Evaluate function at u
        fu = linef.f(u);
        // Reorganise bracket
        if (fu <= fx) {
            if (u >= x)
                br_min = x;
            else
                br_max = x;
            v = w;
            w = x;
            x = u;
            fv = fw;
            fw = fv;
            fx = fu;
        } else {
            if (u < x)
                br_min = u;   
            else
                br_max = u;
            if (fu <= fw || w == x) {
                v = w; 
                w = u;
                fv = fw;
                fw = fu;
            } else if ((fu <= fv) || (v == x) || (v == w)) {
                v = u;
                fv = fu;
            }
        }
        return x;
    }
    

}